# Redis driver
# please install redis-py
#
import asyncio
import time
from datetime import timedelta
from functools import wraps
from typing import Any, Optional, Protocol, Union
from uuid import uuid4

from jcramda import first
from redis.asyncio import Redis
from redis.asyncio.cluster import RedisCluster

from jcutil.core import get_running_loop

__clients = {}


class RedisClientP(Protocol):
    async def exists(self, key) -> bool: ...
    async def get(self, key) -> Any: ...
    async def set(self, key, value, ex: Optional[int] = None) -> bool: ...
    async def setnx(self, key, value) -> int: ...
    async def delete(self, key) -> int: ...
    async def expire(self, key, seconds) -> int: ...


async def new_client(uri: str, tag: str = None):
    if tag is None:
        tag = uuid4()

    if uri.startswith("cluster"):
        __clients[tag] = RedisCluster.from_url(uri)
    else:
        __clients[tag] = Redis.from_url(uri)


def get_client(tag: Union[str, int] = None) -> RedisClientP:
    """
    获取redis客户端
    """
    if tag is None:
        tag = first(list(__clients.keys()))
    pool = __clients.get(tag, None)
    if pool:
        return pool
    raise RuntimeWarning(f"Not found redis client with name: {tag}")


def load(conf: dict):
    """
    一次性读取配置文件，生成redis链接
    配置文件格式：dict(redis_uri="{redis_uri}")
    ```
    {
      "redis1": "redis://localhost:6379",
    }
    """
    if conf and len(conf) > 0:
        try:
            # 检查是否在异步环境中运行
            loop = asyncio.get_running_loop()

            # 已经在异步环境中，直接创建任务
            async def _load_clients():
                tasks = []
                for key in conf:
                    tasks.append(new_client(conf[key], key))
                await asyncio.gather(*tasks)

            # 在已有异步环境中，返回协程给调用者处理
            return _load_clients()
        except RuntimeError:
            # 不在异步环境中，创建新的事件循环
            loop = get_running_loop()
            tasks = [loop.create_task(new_client(conf[key], key)) for key in conf]
            loop.run_until_complete(asyncio.gather(*tasks))

            # print(f'redis: {key} connected')


connect = conn = get_client


class Lock:
    def __init__(self, tag, lock_flg):
        self._client = connect(tag)
        self._flg = lock_flg
        self._owner = False

    async def acquire(self, blocking=True, timeout=None):
        if not blocking and timeout is not None:
            raise ValueError("can't specify timeout for non-blocking acquire")
        start_time = int(time.time())
        while await self._client.exists(self._flg):
            if not blocking:
                break
            await asyncio.sleep(0.5)
            curr_time = int(time.time())
            if isinstance(timeout, int) and curr_time - start_time > timeout:
                return False
        r = await self._client.setnx(self._flg, int(time.time()))
        if r:
            self._owner = True
            return True
        return False

    async def release(self):
        if self._owner:
            await self._client.delete(self._flg)
            self._owner = False

    @property
    def locked(self):
        return self._owner


class SpinLock(Lock):
    def __init__(self, tag, flag, blocking=True, timeout=None):
        super(SpinLock, self).__init__(tag, flag)
        self._blocking = blocking
        self._timeout = timeout

    async def __aenter__(self):
        r = await self.acquire(self._blocking, self._timeout)
        if not r:
            raise RuntimeError(f"Cannot acquire lock: {self._flg}")
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        await self.release()


class IntervalLimitError(RuntimeError):
    pass


def interval_lock(flg, name, timeout=timedelta(seconds=3)):
    def limit_ann(fn):
        @wraps(fn)
        async def wrapped(*args, **kwargs):
            client = connect(flg)
            time_limit = int(timeout.total_seconds())
            if await client.exists(name):
                raise IntervalLimitError(f"time interval limit in {time_limit}s")
            await client.setnx(name, time.time())
            await client.expire(name, time_limit)
            return await fn(*args, **kwargs)

        return wrapped

    return limit_ann



from enum import IntEnum
from typing import AnyStr, Tuple

BS: int

Model: IntEnum

def aes_encrypt(key: AnyStr, plain: str, /, mode: Model = ...) -> Tuple[str, str]: ...
def aes_decrypt(
    key: AnyStr, value: str, /, mode: Model = ..., nonce_or_iv: str = ...
) -> str: ...
def get_sha1prng_key(key: str) -> str: ...
def aes_ecb_encrypt(key: AnyStr, plaintext: str) -> str: ...
def aes_ecb_decrypt(key: AnyStr, ciphertext: str) -> str: ...
def aes_cfb_encrypt(key: AnyStr, plaintext: str) -> Tuple[str, str]: ...
def aes_cfb_decrypt(key: AnyStr, ciphertext: str, /, iv: str) -> str: ...
def to_base64(hexstring: str) -> str: ...
def sha3sum(plaintext: str) -> str: ...



import time

import pandas as pd

from jcutil.drivers import db


def quick_update(schema, target: str, set_cols, where_cols, data, /, mode="update"):
    """利用pandas快速生成的临时表来批量更新数据

    **使用RPLACE模式需要注意，如果传入的数据字段少于原有字段，则原表中的这些字段将被至为NULL**

    Args:
        schema (str): Database bind alias name
        target (str): update target table name
        set_cols (str, list[str]): Will be updated columns. 'id,name,somecol' or ['id','name','somecol']
        where_cols (str, list[str]): Columns that be used in where. 'id,name,somecol' or ['id', 'name', 'somecol']
        data (Iterable, DataFrame): a data that can pass to pd.DataFrame
        mode (str, optional): sql mode: ['update', 'replace']. Defaults to 'update'.

    Exceptions:
        ValueError: some column in set_cols or where_cols which is not in target table.
    """
    temp_name = f"tmp_{int(time.time())}"
    if isinstance(where_cols, (str, bytes)):
        where_cols = where_cols.split(",")
    if isinstance(set_cols, (str, bytes)):
        set_cols = set_cols.split(",")
    target_columns = set(
        pd.read_sql(f"describe {target}", db.conn(schema)).Field.values
    )
    df = pd.DataFrame(data)
    if diff := (set(set_cols) - target_columns):
        raise ValueError(f"set_cols: {diff} columns not in target table.")
    if diff := (set(where_cols) - target_columns):
        raise ValueError(f"where_cols: {diff} columns not in target table.")
    cols = [col for col in df.columns if col in target_columns]

    if df.empty or not cols:
        return
    if mode == "update":
        sql = "UPDATE {target} t, {source} s SET {sets} WHERE {where}".format(
            target=target,
            source=temp_name,
            sets=",".join([f"t.{c} = s.{c}" for c in set_cols]),
            where=" AND ".join([f"t.{c} = s.{c}" for c in where_cols]),
        )
    else:
        sql = "REPLACE INTO {target} ({sets}) SELECT {sets} FROM {source}".format(
            target=target, source=temp_name, sets=",".join(cols)
        )
    with db.conn(schema).begin() as session:
        session.execute(f"create table {temp_name} like {target}")
        df[cols].to_sql(temp_name, session, index=False, if_exists="append")
        session.execute(sql)
        session.execute(f"drop table {temp_name}")
        session.commit()



import os

import pytest
from jcramda import hex_token, members

from jcutil.crypto import Mode, aes_decrypt, aes_encrypt, sha3sum, to_base64


@pytest.fixture()
def modes():
    return members(Mode)


def test_aes(modes):
    raw = os.getenv("TXT", "This is a Test text")
    print("plaintext =", raw, "| hashsum:", sha3sum(raw))
    # genrate a 32 bytes key
    key = hex_token(32)
    print("key:", key)
    for m in modes:
        print(m)
        ciphertext, iv = aes_encrypt(key, raw, mode=m)
        print(ciphertext, iv)
        r = aes_decrypt(key, mode=m, nonce_or_iv=iv)(ciphertext)
        assert r == raw
        print(to_base64(ciphertext))



from datetime import timedelta
from typing import Any, Callable, NoReturn, TypeVar, Union

from jcutil.defines import Writable

T = TypeVar("T")

def mem_cache(cache_dir: str = ...) -> Callable: ...
def clear_mem(path: str = ..., cache_dir: str = ...) -> NoReturn: ...
def redis_cache(
    expires: Union[int, timedelta],
    prefix: str = ...,
    redis_connector: Callable = ...,
    result_assert: Callable[[Any], bool] = ...,
) -> Callable[[Callable], Any]: ...
def persistence(
    fs: Union[Writable, Callable[[], Writable]], f: Callable[..., T]
) -> Callable[..., T]: ...



import logging
from importlib import import_module
from typing import Union

from jcramda import loc

__all__ = [
    "connect",
    "conn",
    "init_engine",
]

__engines = dict()


def init_engine(tag: str, *args, create_engine=None, **kwargs):
    """Init a oracle engine

    :param tag: a flag with connection engine
    :param create_engine: a function create a database connection engine
    :param kwargs: database connection config

    **Options**
     * schema: default is 'oracle', can use ['oracle', 'mysql', 'sqlite', 'postgres', etc...]
     * user: database connection username
     * password: database connection password
     * dsn: database host and port dsn string
     * url: database connection url, like that: "oracle://username:password@10.0.0.5:1521/sid?encoding=utf-8

    :return:
    """
    schema = kwargs.get("schema", "oracle")
    if create_engine is None:
        sqlmodule = import_module("sqlalchemy")
        url = (
            "{schema}://{user}:{password}@{dsn}?encoding=utf-8".format(
                schema=schema,
                user=kwargs["user"],
                password=kwargs["password"],
                dsn=kwargs["dsn"],
            )
            if "url" not in kwargs
            else kwargs.pop("url")
        )
        current_engine = sqlmodule.create_engine(
            url, pool_size=10, encoding="utf-8", **kwargs
        )
    else:
        current_engine = create_engine(*args, **kwargs)
    __engines[tag] = current_engine
    return current_engine


def new_client(tag, *args, create_engine=None, **kwargs):
    if create_engine is None:
        try:
            module = import_module("sqlalchemy")
            create_engine = getattr(module, "create_engine")
        except ModuleNotFoundError as err:
            logging.error(err)

    __engines[tag] = create_engine(*args, **kwargs)
    return __engines[tag]


def get_client(name: Union[str, int] = 0):
    if len(__engines) > 0:
        return loc(name, __engines)
    raise RuntimeError("no any host can connect.")


def connect(n: Union[str, int] = 0):
    return get_client(n).connect()


def load(conf: dict):
    """
    一次性读取配置文件，生成数据库链接
    配置文件格式：dict(dbname="{dburl}")
    ```
    {
      "db1": "mysql://username:password@127.0.0.1:3306/dbname?encoding=utf8mb",
      "myoracle": "oracle://...",
    }
    ```
    @param conf: Dict[str, str]
    """
    if conf and len(conf) > 0:
        for key in conf:
            try:
                init_engine(key, url=conf[key])
            except Exception as err:
                logging.warning(f"load database [{key}] failed: {err}")
            # print(f'database [{key}] connected')


conn = connect


def instances():
    return [*__engines.keys()]
